-- String routines (tested with Euphoria 2.4)
--
-- (c) Juergen Luethje <http://home.arcor.de/luethje/>
-- mbox 1.00 -- 2003, Nov 09
-- Freeware
-- License see "mbox_eng.txt"


global type string (object x)
   -- called by verify(), rverify(), remove_any(), crunch(),
   --   trim_left_chars(), trim_right_chars(), trim_chars(),
   --   trim_left(), trim_right(), trim(), trim_DQ(), parse_any(),
   --   mail_date(), find_field(), find_field_char(), parse_header(),
   --   mail_header(), header_string(), decode_QPencoded_text(),
   --   decode_header(), From_line(), trim_subject(),
   --   subject_to_filename(), Subj_Date(), parse_mbox(),
   --   set_file_time(), del_file()
   object t

   if atom(x) then return 0 end if
   for i = 1 to length(x) do
      t = x[i]
      if (not integer(t)) or (t < 0)  or (t > #FF) then
         return 0
      end if
   end for
   return 1
end type


global function matchs (sequence s, sequence source, integer start)
   -- searches for s in source, starting at start
   -- in : s      : must not be an empty sequence
   --      source : empty sequence ==> function returns 0
   --      start  : [1, length(source)]
   -- called by replace(), decode_header()
   integer p

   if start >= 1 and start <= length(source) then
      p = match(s, source[start..length(source)])
      if p then
         return p + start - 1
      end if
   end if
   return 0
end function


global function finds (object x, sequence source, integer start)
   -- searches for x in source, beginning at start
   -- in : x      : any object
   --      source : empty sequence ==> function returns 0
   --      start  : [1, length(source)]
   -- called by parse_any(), find_field()
   integer p

   if start >= 1 and start <= length(source) then
      p = find(x, source[start..length(source)])
      if p then
         return start + p - 1
      end if
   end if
   return 0
end function


function finds_any (sequence s, sequence source, integer start)
   -- searches for any element of s in source, beginning at start;
   -- returns the position of the first match, or 0 if nothing found
   -- in:  s      : empty sequence ==> function returns 0
   --      source : empty sequence ==> function returns 0
   --      start  : [1, length(source)]
   -- called by parse_any()

   if start >= 1 then
      for i = start to length(source) do
         if find(source[i], s) then
            return i
         end if
      end for
   end if
   return 0
end function


global function verify (integer start, string source, string charlist)
   -- searches for the first character in source, that is not contained in charlist;
   -- returns the position, or 0 if nothing found
   -- in : start    : [1, length(source)]
   --      source   : empty sequence ==> function returns 0
   --      charlist : empty sequence ==> function returns 0
   -- called by trim_left_chars(), parse_any(), find_field_char(),
   --   decode_header(), From_line()

   if start >= 1 and length(charlist) then
      for i = start to length(source) do
         if not find(source[i], charlist) then
            return i
         end if
      end for
   end if
   return 0
end function


function rverify (integer start, string source, string charlist)
   -- called by trim_right_chars()

   if start < 0 then
      start = length(source)
   elsif start > length(source) then
      return 0
   end if

   for i = start to 1 by -1 do
      if not find(source[i], charlist) then
         return i
      end if
   end for
   return 0
end function


global function replace (sequence source, sequence WhatToReplace, sequence ReplaceWith)
   -- called by ProcessFile()
   integer posn

   if length(WhatToReplace) then
      posn = match(WhatToReplace, source)
      while posn do
         source = source[1..posn-1] & ReplaceWith & source[posn+length(WhatToReplace)..length(source)]
         posn = matchs(WhatToReplace, source, posn+length(ReplaceWith))
      end while
   end if
   return source
end function


function remove_any (string source, string charlist)
   -- called by parse_any()
   sequence ret
   integer pntr, char

   pntr = 0
   ret = repeat(0, length(source))
   for i = 1 to length(source) do
      char = source[i]
      if not find(char, charlist) then
         pntr += 1
         ret[pntr] = char
      end if
   end for
   return ret[1..pntr]
end function


global function crunch (string source, string charlist, integer newchar)
   -- called by subject_to_filename()
   sequence ret
   integer pntr, char, flag

   pntr = 0
   flag = 1
   ret = repeat(0, length(source))
   for i = 1 to length(source) do
      char = source[i]
      if not find(char, charlist) then
         pntr += 1
         ret[pntr] = char
         flag = 1
      elsif flag then
         pntr += 1
         ret[pntr] = newchar
         flag = 0
      end if
   end for
   return ret[1..pntr]
end function


--======================================================================

function trim_left_chars (string source, string charlist)
   -- called by trim_chars(), trim_left()
   integer p

   p = verify(1, source, charlist)
   if p then
      return source[p..length(source)]
   else
      return ""
   end if
end function


function trim_right_chars (string source, string charlist)
   -- called by trim_chars(), trim_right()
   integer p

   p = rverify(-1, source, charlist)
   if p then
      return source[1..p]
   else
      return ""
   end if
end function


function trim_chars (string source, string charlist)
   -- called by trim()
   return trim_left_chars(trim_right_chars(source, charlist), charlist)
end function


constant WHITESPACE = {'\t', ' ', 10, 11, 12, 13}


global function trim_left (string source)
   -- called by header_string(), trim_subject()
   return trim_left_chars(source, WHITESPACE)
end function


global function trim_right (string source)
   -- called by parse_header(), trim_subject()
   return trim_right_chars(source, WHITESPACE)
end function


global function trim (string source)
   -- called by subject_to_filename()
   return trim_chars(source, WHITESPACE)
end function


--======================================================================

constant DQ = '"'

global function trim_DQ (string source)
   -- called by mail_date()
   integer n

   n = length(source)
   if n >= 2 and source[1] = DQ and source[n] = DQ then
      return source[2..n-1]
   end if
   return source
end function

global function parse_any (string source, string delimlist)
   -- split a string into "words"
   -- in : source    : empty sequence ==> function returns {}
   --      delimlist : empty sequence ==> function returns {}
   -- called by mail_date(), decode_header()
   sequence ret
   integer begpos, p, dq

   delimlist = remove_any(delimlist, {DQ})
   ret = {}
   p = 1
   dq = 0
   while 1 do
      begpos = verify(p+dq, source, delimlist)
      if begpos = 0 then exit end if

      if source[begpos] = DQ then
         dq = 1
         p = finds(DQ, source, begpos+1)
      else
         dq = 0
         p = finds_any(delimlist, source, begpos)
      end if
      if p then
         ret = append(ret, source[begpos..p-1+dq])
      else
         ret = append(ret, source[begpos..length(source)])
         exit
      end if
   end while
   return ret
end function

