-- Date and time operations (tested with Euphoria 2.4)
--
-- (c) Juergen Luethje <http://home.arcor.de/luethje/>
-- mbox 1.00 -- 2003, Nov 09
-- Freeware
-- License see "mbox_eng.txt"

include math.e
include string.e


global constant
   DAYS_SHORT    = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"},
   MONTHS_SHORT  = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
    SECS_PER_DAY = 86400,      --      seconds per day
   MSECS_PER_DAY = 86400000,   -- milliseconds per day

   -- The following constants are used to refer to the elements of
   -- EUPHORIA's date() function:
   E_YEAR = 1, E_MONTH = 2, E_DAY = 3,
   E_HOUR = 4, E_MINUTE = 5, E_SECOND = 6


function leap_year (integer year)
   -- out: TRUE/FALSE
   -- called by days_in_month()
   if (remainder(year, 4) = 0) and
     ((remainder(year, 100) != 0) or (remainder(year, 400) = 0)) then
      return 1
   end if
   return 0
end function


constant DaysInMonth = {31,28,31,30,31,30,31,31,30,31,30,31}

function days_in_month (integer year, integer month)
   -- called by tDate(), serial_to_date()
   integer ret

   if month < 1 or month > 12 then return 0 end if
   ret = DaysInMonth[month]
   if (month = 2) and leap_year(year) then ret += 1 end if
   return ret
end function


integer MinYear
MinYear = 1

global type tDate (object x)
   -- {year, month, day}
   -- called by DateTime(), date_serial()
   object year, day

   if atom(x) or length(x) != 3 then
      return 0
   end if

   year = x[1]
   if not integer(year) or year < MinYear or year > 9999 then
      return 0
   end if
   -- x[2] is checked by days_in_month()
   day = x[3]
   if not integer(day) or day < 1 or day > days_in_month(year, x[2]) then
      return 0
   end if

   return 1
end type


global type tTime (object x)
   -- {hour, minute, second}
   -- called by DateTime(), time_serial()
   object hour, minute, second

   if atom(x) or length(x) != 3 then
      return 0
   end if

   hour = x[1]
   if not integer(hour) or hour < 0 or hour > 23 then
      return 0
   end if
   minute = x[2]
   if not integer(minute) or minute < 0 or minute > 59 then
      return 0
   end if
   -- number of seconds allowing for a leap second
   second = x[3]
   if (not atom(second)) or (second < 0) or (60 < second) then
      return 0
   end if

   return 1
end type


global type DateTime (object x)
   -- {year, month, day, hour, minute, second} >= 0001-01-01, 00:00:00,
   -- called by tFileTime()
   if atom(x) or length(x) != 6 then
      return 0
   end if

   if not tDate(x[1..E_DAY]) then
      return 0
   end if
   if not tTime(x[E_HOUR..6]) then
      return 0
   end if

   return 1
end type


global type tFileTime (object x)
   -- called by dos_time(), set_file_time(), ProcessFile()
   integer ret

   MinYear = 1980
   ret = DateTime(x)
   MinYear = 1
   return ret
end type


global type tDateSerial (object x)
   -- number of days since the beginning of the calendar,
   -- if always the Gregorian Calendar had been used
   -- + fraction of the day
   -- note:       1 = date_serial({    1,1,1})
   --       3652060 = date_serial({10000,1,1})
   -- called by serial_to_date(), mail_date(), time_to_file_time_WIN32()
   if atom(x) and 1 <= x and x < 3652060 then
      return 1
   end if
   return 0
end type

------------------------------------------------------------------------

global function date_serial (tDate indate)
   -- in : sequence of the form {year, month, day}
   -- out: number of days since the beginning of the calendar,
   --      if always the Gregorian Calendar had been used
   -- called by mail_date(), time_to_file_time_WIN32()
   integer year, month, t

   year  = indate[E_YEAR]
   month = indate[E_MONTH]
   t = (year-1)*365 + (month-1)*31 + indate[E_DAY]
   if month >= 3 then
      t -= floor(0.4*month + 2.3)
   else
      year -= 1
   end if
   return t + floor(year/4) - floor(year/100) + floor(year/400)
end function


function time_serial (tTime intime)
   -- in : sequence of the form {hour, minute, second}
   -- out: fraction of a day
   -- called by mail_date()

   return ((intime[1]*60 + intime[2])*60 + intime[3]) / SECS_PER_DAY
end function


constant
   DAYS1   = 365,                   -- days in   1 year
   DAYS4   = DAYS1 * 4 + 1,         -- days in   4 years (  1461)
   DAYS100 = DAYS4 *25 - 1,         -- days in 100 years ( 36524)
   DAYS400 = DAYS100*4 + 1          -- days in 400 years (146097)

function serial_to_date (tDateSerial dateserial)
   -- in : The integer part of 'dateserial' is the number of days since
   --      the beginning of the calendar, if always the Gregorian
   --      Calendar had been used.
   -- out: sequence of the form {year, month, days}
   -- called by mail_date()

   integer t, year, month, days, count

   -- calculate the year
   t = floor(dateserial) - 1
   count = floor(t / DAYS400)
   year = count*400 + 1
   t -= count*DAYS400                    -- 0 <= t <= DAYS100*4

   count = floor(t / DAYS100)
   days = remainder(t, DAYS100)
   if count = 4 then                     -- 4*100 years are 1 leapday
      count -= 1                         --   less than 400 years
      days = DAYS100
   end if                                -- 0 <= days <= DAYS100 (36524)
   year += count*100

   count = floor(days / DAYS4)
   days = remainder(days, DAYS4)         -- 0 <= days <= DAYS1*4 ( 1460)
   year += count*4

   count = floor(days / DAYS1)
   days = remainder(days, DAYS1)
   if count = 4 then                     -- 4*1 year are 1 leapday
      count -= 1                         --   less than 4 years
      days = DAYS1
   end if                                -- 0 <= days <= 365
   year += count

   -- calculate month and day
   month = 1
   count = 31
   while days >= count do
      days -= count
      month += 1
      count = days_in_month(year, month)
   end while
   days += 1

   return {year, month, days}
end function

------------------------------------------------------------------------

function serial_to_time (atom timeserial)
   -- in : The fractional part of 'timeserial' is the fraction of a day.
   -- out: sequence of the form {hour, minute, second}
   -- called by mail_date()

   integer mSecCount, MinCount, hour, minu, sec

   mSecCount = ceil(frac(timeserial)*MSECS_PER_DAY)
   MinCount = floor(mSecCount / 60000)
   mSecCount = remainder(mSecCount, 60000)
   hour = floor(MinCount / 60)
   minu = remainder(MinCount, 60)
   sec  = floor(mSecCount / 1000)
   return {hour, minu, sec}
end function

------------------------------------------------------------------------

constant                -- obsolete three character US time zones
   OBS_US_ZONES_S = {"PST","MST","CST","EST"},  -- -8, -7, -6, -5
   OBS_US_ZONES_D = {"PDT","MDT","CDT","EDT"}   -- -7, -6, -5, -4

global function mail_date (string maildate)
   -- in : body of an e-mail 'Date' header field, e.g.
   --         [Sat,] 5 Apr 2003 14:33[:24] -0500 [...]
   --         [Sat,] 5 Apr   03 14:33[:24] ["]EST["]
   --      (obsolete format for year and time zone taken into account,
   --       after RFC 2822)
   -- out: date and time in 'DateTime' format
   --      (expressed in *UTC*, if valid time zone was given)
   -- called by Subj_Date();
   -- The return value is used by dos_time(), and set_file_time().
   sequence ret, d, t, zone_s
   integer p, sig
   atom year, zone, minu
   tDateSerial utc

   ret = {-1,-1,-1,-1,-1,0}                 -- The seconds are optional.
   d = parse_any(maildate, ", \t\n\r")
   if length(d) = 0 then return ret end if

   p = 2                                    -- standard
   if find(d[1], DAYS_SHORT) = 0 then
      p = 1                                 -- no weekday given
   end if
   if length(d) < p+4 then return ret end if

   -- local date
   ret[E_DAY] = val(d[p])
   if val_err() then return ret end if           -- invalid day
   ret[E_MONTH] = find(d[p+1], MONTHS_SHORT)
   year = val(d[p+2])
   if year < 0 then                              -- invalid year
      return ret
   elsif year < 50 then
      year += 2000
   elsif year < 1000 then
      year += 1900
   end if
   ret[E_YEAR] = year

   -- local time
   t = parse_any(d[p+3], ":")
   if length(t) < 2 or length(t) > 3 then return ret end if
   for i = 1 to length(t) do
      ret[3+i] = val(t[i])
      if val_err() then       -- invalid time
         ret[3+i] = -1
         return ret
      end if
   end for

   -- time zone
   sig = 0
   zone_s = d[p+4]
   if length(zone_s) = 5 then
      zone = val(zone_s[2..3])                   -- hours
      if (val_err() = 0) and integer(zone) then
         minu = val(zone_s[4..5])                -- minutes
         if (val_err() = 0) and integer(minu) and (minu <= 59) then
            sig = find(zone_s[1], "-+")
            if sig != 0 then
               zone += minu/60
               if sig = 1 then
                  zone = -zone
               end if
            end if
         end if
      end if
   end if
   if sig = 0 then                               -- invalid time zone
      zone_s = trim_DQ(zone_s)
      zone = find(zone_s, OBS_US_ZONES_S) - 9
      if zone = -9 then                          -- not found
         zone = find(zone_s, OBS_US_ZONES_D)
         if zone != 0 then zone -= 8 end if
      end if
   end if
   if zone = 0 then return ret end if   -- Time zone is Universal Time,
                                        -- or given in invalid format.
   -- calc UTC
   utc = date_serial(ret[E_YEAR..E_DAY])
       + time_serial(ret[E_HOUR..E_SECOND]) - zone/24
   return serial_to_date(utc) & serial_to_time(utc)
end function


global function dos_time (sequence indate)
   -- in: {year,month,day,hour,minute,second}
   -- called by ReadHeader()
   if tFileTime(indate) and indate[1] <= 2100 then    -- year between 1980 and 2100
      indate[1] -= 1980
      return (indate[1]*#200 + indate[2]*#20 + indate[3]) * #10000  -- date
            + indate[4]*#800 + indate[5]*#20 + floor(indate[6]/2)   -- time
   else
      return #210000   -- 1. Jan. 1980, 00:00:00
   end if
end function

